import asyncio
import logging

from nyuki.api import Api
from nyuki.services import Service


log = logging.getLogger(__name__)


def resource(endpoint, version=None, content_type=None):
    """
    Nyuki resource decorator to register a route.
    A resource has multiple HTTP methods (get, post, etc).
    """
    def decorated(cls):
        cls.endpoint = endpoint
        cls.version = version
        cls.content_type = content_type or 'application/json'
        return cls
    return decorated


class Capability(object):

    """
    A capability is unique (hashable object, based on capability's name).
    """

    def __init__(self, name, method, endpoint, version, handler, wrapper):
        self.name = name
        self.method = method
        self.endpoint = endpoint
        self.version = version
        self.handler = handler
        self.wrapper = wrapper

    def __hash__(self):
        return hash(self.name)


class Exposer(Service):

    CONF_SCHEMA = {
        "type": "object",
        "required": ["api"],
        "properties": {
            "api": {
                "type": "object",
                "properties": {
                    "host": {"type": "string"},
                    "port": {"type": "integer"}
                }
            }
        }
    }

    def __init__(self, nyuki):
        self._nyuki = nyuki
        self._nyuki.register_schema(self.CONF_SCHEMA)
        self._loop = self._nyuki.loop or asyncio.get_event_loop()
        self._capabilities = set()
        self._api = Api(self._loop)
        self.host = None
        self.port = None

    async def start(self):
        """
        Expose capabilities by building the HTTP server.
        The server will be started with the event loop.
        """
        await self._api.build(self.host, self.port)

    def configure(self, host='0.0.0.0', port=5558, debug=False):
        self.host = host
        self.port = port
        self._api.debug = debug

    async def stop(self):
        await self._api.destroy()

    @property
    def capabilities(self):
        return self._capabilities

    def register(self, capa):
        """
        Add a capability and register its route to the API.
        """
        if capa in self._capabilities:
            raise ValueError("A capability is already exposed through {} with "
                             "{} method".format(capa.endpoint, capa.method))
        self._capabilities.add(capa)
        # Handle version as part of the endpoint (if defined)
        if capa.version:
            endpoint = '/{}{}'.format(capa.version, capa.endpoint)
        else:
            endpoint = capa.endpoint
        self._api.router.add_route(capa.method, endpoint, capa.wrapper)
        log.debug("Capability added: {}".format(capa.name))

    def _find(self, capa_name):
        """
        Get a capability by its name.
        """
        for capability in self._capabilities:
            if capa_name == capability.name:
                return capability

    def call(self, name, request):
        """
        Call a capability by its name in an asynchronous fashion.
        """
        capa = self._find(name)
        if not capa:
            log.warning("Capability {} is called but doen't exist".format(name))
            return

        # TODO: use asyncio.ensure_future when Python 3.4.4 will be released
        future = asyncio.ensure_future(capa.wrapper(request), loop=self._loop)
        return future



import asyncio
from datetime import datetime
import json
import logging
from slixmpp import ClientXMPP
from slixmpp.exceptions import IqError, IqTimeout
from uuid import uuid4

from nyuki.bus import reporting
from nyuki.bus.persistence import BusPersistence, EventStatus, PersistenceError
from nyuki.services import Service


log = logging.getLogger(__name__)


class PublishError(Exception):
    pass


class _BusClient(ClientXMPP):

    """
    XMPP client to connect to the bus.
    This class is based on Slixmpp (fork of Sleexmpp) using asyncio event loop.
    """

    def __init__(self, jid, password, host=None, port=5222, certificate=None):
        super().__init__(jid, password)
        host = host or self.boundjid.domain
        self._address = (host, port)

        self.register_plugin('xep_0045')  # Multi-user chat
        self.register_plugin('xep_0077')  # In-band registration

        # Disable IPv6 until we really need it
        self.use_ipv6 = False

        if certificate:
            self.ca_certs = certificate

    def connect(self, **kwargs):
        """
        Connect to the XMPP server using the default address computed at init
        if not passed as argument.
        """
        try:
            self._address = kwargs.pop('address')
        except KeyError:
            pass
        super().connect(address=self._address, **kwargs)
        self.init_plugins()

    def start_tls(self):
        try:
            return super().start_tls()
        except FileNotFoundError:
            log.error('SSL certificates missing at %s', self.ca_certs)
            self.abort()

    def exception(self, exc):
        """
        xmlstream.py catches exceptions itself, calls this method afterwards
        """
        reporting.exception(exc)


class Bus(Service):

    """
    A simple class that implements Publish/Subscribe-like communications over
    XMPP. This communication layer is also called "bus".
    Each nyuki has its own "topic" (actually a MUC) to publish events. And it
    can publish events to that topic only. In the meantime, each nyuki can
    subscribe to any other topics to process events from other nyukis.
    """

    CONF_SCHEMA = {
        "type": "object",
        "required": ["bus"],
        "properties": {
            "bus": {
                "type": "object",
                "required": ["jid", "password"],
                "properties": {
                    "certificate": {
                        "type": "string",
                        "minLength": 1
                    },
                    "host": {
                        "type": "string",
                        "minLength": 1
                    },
                    "jid": {
                        "type": "string",
                        "minLength": 1
                    },
                    "muc_domain": {"type": ["string", "null"]},
                    "password": {
                        "type": "string",
                        "minLength": 1
                    },
                    "port": {"type": "integer"},
                    "persistence": {
                        "type": "object",
                        "properties": {
                            "backend": {
                                "type": "string",
                                "minLength": 1
                            }
                        }
                    }
                },
                "additionalProperties": False
            }
        }
    }

    def __init__(self, nyuki):
        self._nyuki = nyuki
        self._nyuki.register_schema(self.CONF_SCHEMA)
        self._loop = self._nyuki.loop or asyncio.get_event_loop()

        self.client = None

        # Connectivity stuff
        self.reconnect = False
        self._connected = asyncio.Event()
        self._persistence = None
        self._disconnection_datetime = None
        self._publish_futures = dict()

        # MUCs
        self._callbacks = dict()
        self._topic = None
        self._mucs = None
        self._muc_domain = None

    async def start(self, timeout=0):
        self.client.connect()
        if timeout:
            await asyncio.wait_for(self._connected.wait(), timeout)

        try:
            await self._persistence.init()
        except PersistenceError:
            log.error(
                'Could not init persistence storage with %s',
                self._persistence.backend
            )

    def configure(self, jid, password, host='localhost', port=5222,
                  muc_domain='mucs.localhost', certificate=None,
                  persistence={}):
        # XMPP client and main handlers
        self.client = _BusClient(
            jid, password, host, port, certificate=certificate
        )
        self.client.loop = self._loop
        self.client.add_event_handler('connection_failed', self._on_failure)
        self.client.add_event_handler('disconnected', self._on_disconnect)
        self.client.add_event_handler('message', self._on_direct_message)
        self.client.add_event_handler('register', self._on_register)
        self.client.add_event_handler('session_start', self._on_start)
        self.client.add_event_handler('stream_error', self._on_stream_error)

        # MUCs
        self._topic = self.client.boundjid.user
        self._callbacks[self._topic] = None
        self._mucs = self.client.plugin['xep_0045']
        self._muc_domain = muc_domain
        self.client.add_event_handler('groupchat_message', self._on_event)
        self.client.add_event_handler(
            'muc::{}::message_error'.format(self._muc_address(self._topic)),
            self._on_failed_publish
        )

        # Persistence storage
        self._persistence = BusPersistence(name=self._topic, **persistence)
        log.info('Bus persistence set to %s', self._persistence.backend)

    async def stop(self):
        if not self.client:
            return

        if not self.client.transport:
            log.warning('XMPP client is already disconnected')
            return

        self.reconnect = False
        self.client.disconnect(wait=2)

        try:
            await asyncio.wait_for(self.client.disconnected, 2.0)
        except asyncio.TimeoutError:
            log.error('Could not end bus connection after 2 seconds')

        await self._persistence.close()

    def _muc_address(self, topic):
        return '{}@{}'.format(topic, self._muc_domain)

    async def _on_register(self, event):
        """
        XMPP event handler while registering a user (in-band registration).
        """
        resp = self.client.Iq()
        resp['type'] = 'set'
        resp['register']['username'] = self.client.boundjid.user
        resp['register']['password'] = self.client.password

        try:
            await resp.send()
        except IqError as exc:
            error = exc.iq['error']['text']
            log.debug("Could not register account: {}".format(error))
        except IqTimeout:
            log.error("No response from the server")
        finally:
            log.debug("Account {} created".format(self.client.boundjid))

    async def _on_start(self, event):
        """
        XMPP event handler when the connection has been made.
        """
        log.info('Connected to XMPP server at {}:{}'.format(
            *self.client._address
        ))
        self.reconnect = True
        self.client.send_presence()
        self.client.get_roster()

        # Auto-subscribe to the topic where the nyuki could publish events.
        self._connected.set()
        await self.subscribe(self._topic)

        # Replay events that have been lost, if any
        if self._persistence and self._disconnection_datetime:
            await self.replay(
                self._disconnection_datetime, EventStatus.not_sent(), 3
            )
        self._disconnection_datetime = None

    async def _on_disconnect(self, event):
        """
        XMPP event handler when the client has been disconnected.
        """
        if self._connected.is_set():
            log.warning('Disconnected from the bus')
            self._connected.clear()

        self._disconnection_datetime = datetime.utcnow()

        if self.reconnect:
            # Restart the connection loop
            await self.client._connect_routine()

    def _on_failure(self, event):
        """
        XMPP event handler when something is going wrong with the connection.
        """
        log.error('Connection to XMPP server {}:{} failed'.format(
            *self.client._address
        ))
        # If prosody was not up at nyuki started
        if not self._disconnection_datetime:
            self._disconnection_datetime = datetime.utcnow()
        self.client.abort()

    async def _on_stream_error(self, event):
        """
        Received on stream_error from prosody
        (TODO: event informations unknown)
        """
        log.error('Received unexpected stream error, resubscribing')
        log.debug('event dump: %s', event)
        await self._resubscribe()
        if event['id'] in self._publish_futures:
            if not self._publish_futures[event['id']].done():
                self._publish_futures[event['id']].set_exception(PublishError)
            else:
                log.debug('Publish future was already done')

    async def _on_failed_publish(self, event):
        """
        Event received when a publish has failed
        (publish informations in the event)
        """
        uid = event['id']
        muc = event['from'].user
        log.warning("Was not subscribed to MUC '%s', retrying", muc)
        await self.subscribe(muc, self._callbacks.get(muc))
        if not self._publish_futures[uid].done():
            self._publish_futures[uid].set_exception(PublishError)
        else:
            log.debug('Publish future was already done for id %s', uid)

    async def _on_event(self, event):
        """
        XMPP event handler when a Nyuki Event has been received.
        """
        # ignore events from the nyuki itself
        efrom = event['from'].user
        if efrom == self._topic:
            future = self._publish_futures.get(event['id'])
            if not future:
                log.warning('Received own publish that was not in memory')
            else:
                future.set_result(None)
            return

        log.debug('event received: {}'.format(event))

        try:
            body = json.loads(event['body'])
        except ValueError:
            body = {}

        callback = self._callbacks.get(efrom)
        if callback:
            log.debug('calling callback %s', callback)
            if not asyncio.iscoroutinefunction(callback):
                log.warning('event callbacks must be coroutines')
                callback = asyncio.coroutine(callback)
            await callback(body)
        else:
            log.warning('No callback set for event from %s', efrom)

    async def replay(self, since=None, status=None, wait=0):
        """
        Replay events since the given datetime (or all if None)
        """
        log.info('Replaying events')
        if since:
            log.info('    since %s', since)
        if status:
            log.info('    with status %s', status)

        if wait:
            log.info('Waiting %d seconds', wait)
            await asyncio.sleep(wait)

        events = await self._persistence.retrieve(since, status)
        for event in events:
            await self.publish(
                json.loads(event['message']),
                event['topic'],
                event['id']
            )

    async def publish(self, event, dest=None, previous_uid=None):
        """
        Send an event in the nyuki's own MUC so that other nyukis that joined
        the MUC can process it.
        """
        if self._muc_domain is None:
            log.warning('No subscription to any muc')
            return

        if not isinstance(event, dict):
            raise TypeError('Message must be a dict')

        msg = self.client.Message()
        msg['id'] = uid = previous_uid or str(uuid4())
        msg['type'] = 'groupchat'
        msg['to'] = self._muc_address(dest or self._topic)
        msg['body'] = json.dumps(event)

        self._publish_futures[uid] = asyncio.Future()

        # Publish in MUC
        if self._connected.is_set():
            log.debug(">> publishing to '{}': {}".format(self._topic, event))
            log.info('Publishing an event to %s', msg['to'])
            status = None
            while status != EventStatus.SENT:
                msg.send()
                try:
                    await self._publish_futures[uid]
                except PublishError:
                    status = EventStatus.FAILED
                    self._publish_futures[uid] = asyncio.Future()
                else:
                    status = EventStatus.SENT
                    log.info("Event successfully sent to MUC '%s'", msg['to'])
        else:
            status = EventStatus.PENDING

        del self._publish_futures[uid]

        # Once we have a result, store it if needed
        if previous_uid:
            await self._persistence.update(previous_uid, status)
        elif self._persistence:
            await self._persistence.store({
                'id': uid,
                'status': status.value,
                'topic': dest or self._topic,
                'message': msg['body'],
            })
            in_memory = self._persistence.memory_buffer
            if in_memory.is_full:
                asyncio.ensure_future(
                    self._nyuki.on_buffer_full(in_memory.free_slot)
                )

    async def _resubscribe(self):
        """
        Resubscribe everywhere
        """
        for topic in self._callbacks.keys():
            await self.subscribe(topic, self._callbacks.get(topic))

    async def subscribe(self, topic, callback=None):
        """
        Enter into a new chatroom using xep_0045.
        Room address format must be like '{topic}@mucs.localhost'
        """
        if not self._connected.is_set():
            log.warning("Waiting for a connection to subscribe to '%s'", topic)
        await self._connected.wait()

        self._mucs.joinMUC(self._muc_address(topic), self._topic)
        log.info("Subscribed to '{}'".format(topic))
        if topic not in self._callbacks:
            self._callbacks[topic] = callback
        else:
            log.warning("Callback already set for topic: %s", topic)

    async def unsubscribe(self, topic):
        """
        Leave a chatroom.
        """
        if not self._connected.is_set():
            log.warning("Waiting for a connection to unsubscribe from '%s'", topic)
        await self._connected.wait()

        self._mucs.leaveMUC(self._muc_address(topic), self._topic)
        del self._callbacks[topic]
        log.info("Unsubscribed from '{}'".format(topic))

    def direct_subscribe(self, callback):
        """
        Enable handling of direct messages from other nyukis
        """
        if not asyncio.iscoroutinefunction(callback):
            log.warning('event callbacks must be coroutines')
            callback = asyncio.coroutine(callback)
        self._direct_message_callback = callback
        log.info("Subscribed to direct messages")

    def direct_unsubscribe(self, source):
        """
        Remove the callback for direct messages
        """
        self._direct_message_callback = None
        log.info('Unsubscribed from direct messages')

    async def _on_direct_message(self, event):
        """
        Direct message events end up here
        """
        if event.get('type') == 'groupchat':
            return

        log.debug('direct message received: {}'.format(event))
        efrom = event['from'].user

        try:
            body = json.loads(event['body'])
        except ValueError:
            body = {}

        if self._direct_message_callback:
            log.debug('calling direct message callback')
            await self._direct_message_callback(efrom, body)
        else:
            log.warning('No callback set for direct messages')

    async def send_message(self, recipient, data):
        """
        Send a direct message to 'recipient'
        """
        if not recipient:
            log.debug('No recipient for direct message, ignoring it')
            return
        if not isinstance(data, dict):
            raise TypeError('Message must be a dict')
        if not self._connected.is_set():
            log.info("Waiting for a connection to direct message '%s'", recipient)
        await self._connected.wait()
        log.debug(">> direct message to '{}': {}".format(recipient, data))
        body = json.dumps(data)
        self.client.send_message(recipient, body)



import asyncio
import json
from jsonschema import validate, ValidationError
import logging
import random
import re
import string
import websockets

from nyuki.services import Service


log = logging.getLogger(__name__)


def websocket_ready(func):
    """
    This decorator set an async method to be called at a client connection,
    the return value (dict) and will be sent as data to the client
    """
    @staticmethod
    async def decorated(self, token):
        return await func(self, token)

    WebHandler.READY_CALLBACK = decorated
    return func


def websocket_close(func):
    """
    This decorator set an async method to be called at a client connection,
    the return value (dict) and will be sent as data to the client
    """
    @staticmethod
    async def decorated(self, token):
        return await func(self, token)

    WebHandler.CLOSE_CALLBACK = decorated
    return func


class WebHandler(Service):

    CONF_SCHEMA = {
        "type": "object",
        "properties": {
            "websocket": {
                "type": "object",
                "properties": {
                    "host": {"type": "string"},
                    "port": {"type": "integer"}
                }
            }
        }
    }

    READY_CALLBACK = None
    CLOSE_CALLBACK = None
    TOKEN_USAGE_TIMEOUT = 1
    KEEPALIVE_SCHEMA = {
        'type': 'object',
        'required': ['type'],
        'properties': {
            'type': {
                'type': 'string',
                'minLength': 1
            }
        }
    }

    def __init__(self, nyuki, loop=None):
        self._nyuki = nyuki
        self._nyuki.register_schema(self.CONF_SCHEMA)
        self._loop = loop or asyncio.get_event_loop()
        self.host = None
        self.port = None
        self.server = None
        self.keepalive = None
        self.clients = {}

    async def start(self):
        """
        Start the websocket server
        """
        log.info("Starting websocket server on %s:%s", self.host, self.port)
        self.server = await websockets.serve(
            self._wshandler, self.host, self.port
        )

    def configure(self, host='0.0.0.0', port=5559, keepalive=60):
        self.host = host
        self.port = port
        self.keepalive = keepalive

    async def stop(self):
        if self.server:
            for ws in self.server.websockets:
                ws.close()
            self.server.close()
            await self.server.wait_closed()

    def new_token(self):
        """
        Random token using 30 digits/lowercases
        """
        token = ''.join(
            random.choice(string.ascii_letters + string.digits)
            for _ in range(30)
        )
        log.debug('new token: %s', token)
        self.clients[token] = None
        self._loop.call_later(
            self.TOKEN_USAGE_TIMEOUT, self._check_token_usage, token
        )
        return token

    async def broadcast(self, message, tokens=None, timeout=2.0):
        """
        Send a message to every connected client
        """
        tasks = []
        if not isinstance(message, str):
            data = json.dumps(message)
        if isinstance(tokens, str):
            tokens = [tokens]
        if isinstance(tokens, list):
            log.debug('Sending to client list %s: %s', tokens, data)
            for token in tokens:
                client = self.clients[token]
                if client is None:
                    log.warning("Token '%s' initialized but not used", token)
                    continue
                tasks.append(asyncio.ensure_future(client.send(data)))
        else:
            log.debug('Sending to all WS clients: %s', data)
            for websocket in self.server.websockets:
                tasks.append(asyncio.ensure_future(websocket.send(data)))

        if not tasks:
            log.debug('Nobody to broadcast to')
            return

        await asyncio.wait(tasks, timeout=timeout)

    async def _send_ready(self, websocket, token):
        """
        Send the 'ready' message to a client
        """
        self.clients[token] = websocket
        ready = {
            'type': 'ready',
            'keepalive_delay': self.keepalive * 0.8,
            'data': {}
        }
        if self.READY_CALLBACK:
            log.info('Ready callback set up, calling it for new client')
            ready['data'] = await self.READY_CALLBACK(self._nyuki, token) or {}
        log.info('Sending ready packet')
        log.debug('ready dump: %s', ready)
        await websocket.send(json.dumps(ready))

    def _check_token_usage(self, token):
        """
        Delete token if never used
        """
        if token in self.clients and self.clients[token] is None:
            log.debug("token '%s' never used, deleting it", token)
            del self.clients[token]

    async def _end_websocket_client(self, websocket, token, reason):
        """
        Close the connection if no keepalive have been received
        """
        if self.CLOSE_CALLBACK:
            log.info('Close callback set up, calling it before ending client')
            await self.CLOSE_CALLBACK(self._nyuki, token)
        websocket.close(reason=reason)
        try:
            del self.clients[token]
        except KeyError as ke:
            log.debug("token '%s' already removed from keepalive", ke)

    def _schedule_keepalive(self, websocket, token):
        return self._loop.call_later(
            self.keepalive,
            lambda: asyncio.ensure_future(self._end_websocket_client(
                websocket, token, 'keepalive timed out'
            ), loop=self._loop)
        )

    async def _wshandler(self, websocket, path):
        """
        Main handler for a newly connected client
        """
        match = re.match(r'^\/(?P<token>[a-zA-Z0-9]{30})$', path)
        if not match:
            log.debug('token does not match (%s)', path)
            websocket.close()
            return

        token = match.group('token')
        if token not in self.clients:
            log.debug("Unknown token '%s'", token)
            websocket.close()
            return
        elif self.clients[token] is not None:
            log.debug("Token already in use: '%s'", token)
            websocket.close()
            return

        log.info('Connection from token: %s', token)
        handle = self._schedule_keepalive(websocket, token)
        await self._send_ready(websocket, token)

        while True:
            # Main read loop
            try:
                message = await websocket.recv()
            except websockets.exceptions.ConnectionClosed as exc:
                log.debug('client connection closed: %s', exc)
                break

            # Decode JSON message
            try:
                data = json.loads(message)
            except ValueError:
                log.debug('Message received not JSON: %s', message)
                continue

            try:
                validate(data, self.KEEPALIVE_SCHEMA)
            except ValidationError:
                log.debug('Invalid keepalive received: %s', data)
                continue

            mtype = data['type']
            if mtype == 'keepalive':
                handle.cancel()
                handle = self._schedule_keepalive(websocket, token)

        handle.cancel()
        await self._end_websocket_client(
            websocket, token, 'connection closed normally'
        )



import asyncio
from datetime import datetime
from jsonschema import FormatChecker, validate, ValidationError
import logging
import os
import socket
import sys
from traceback import TracebackException

from nyuki.utils import from_isoformat


log = logging.getLogger(__name__)


REPORT_SCHEMA = {
    'type': 'object',
    'required': ['hostname', 'ipv4', 'type', 'author', 'datetime', 'data'],
    'properties': {
        'hostname': {
            'type': 'string',
            'minLength': 1
        },
        'ipv4': {
            'type': 'string',
            'minLength': 1
        },
        'type': {
            'type': 'string',
            'minLength': 1
        },
        'author': {
            'type': 'string',
            'minLength': 1
        },
        'datetime': {
            'type': 'string',
            'format': 'isoformat'
        },
        'data': {'type': 'object'}
    },
    "additionalProperties": False
}


report_checker = FormatChecker()


@report_checker.checks(format='isoformat')
def _check_isoformat(datetime):
    try:
        from_isoformat(datetime)
    except ValueError:
        log.warning('Unknown datetime format: %s', datetime)
        return False
    return True


class Reporter(object):

    EXCEPTION_TTL = 3600

    def __init__(self):
        self.name = None
        self._loop = None
        self._publisher = None
        self._channel = None
        self._handlers = list()
        self._last_exceptions = list()

    def init(self, name, publisher, channel, loop=None):
        if not hasattr(publisher, 'subscribe'):
            raise TypeError("Nyuki publisher requires the 'subscribe' method")
        if not hasattr(publisher, 'publish'):
            raise TypeError("Nyuki publisher requires the 'publish' method")

        self.name = name
        self._loop = loop or asyncio.get_event_loop()
        self._publisher = publisher
        self._channel = channel
        asyncio.ensure_future(
            self._publisher.subscribe(channel, self._handle_report)
        )

    async def _handle_report(self, data):
        """
        Handle report, ignore if it comes from this reporter
        """
        try:
            self.check_report(data)
        except ValidationError:
            log.debug('Received invalid report format, ignoring')
            return

        if data['author'] == self.name:
            log.debug('Received own report, ignoring')
            return

        tasks = []
        for handler in self._handlers:
            tasks.append(asyncio.ensure_future(handler(data)))

        if not tasks:
            log.debug('No report handler to execute')
            return

        await asyncio.wait(tasks)

    def register_handler(self, handler):
        """
        Register all required handlers for received reports
        """
        if handler in self._handlers:
            # Can't register the same callback twice
            return
        self._handlers.append(handler)

    def check_report(self, report):
        """
        Raise ValidationError on failure
        """
        validate(report, REPORT_SCHEMA, format_checker=report_checker)

    def send_report(self, rtype, data):
        """
        Send reports with a type and any data

        Using docker/fleet, we require some informations about IP/hostname of
        our nyuki containers, using environnement vars :
            - MACHINE_NAME: machine hostname
            - DEFAULT_IPV4: local container ipv4
        Otherwise, the nyuki try and search for it by itself.
        """
        if not self._publisher:
            log.warning('Reporting not initiated')
            return

        report = {
            'hostname': os.environ.get('MACHINE_NAME', socket.gethostname()),
            'ipv4': os.environ.get(
                'DEFAULT_IPV4',
                socket.gethostbyname(socket.gethostname())
            ),
            'type': rtype,
            'author': self.name,
            'datetime': datetime.utcnow().isoformat(),
            'data': data
        }
        self.check_report(report)
        log.info("Sending report data with type '%s'", rtype)
        asyncio.ensure_future(self._publisher.publish(report, self._channel))

    def exception(self, exc):
        """
        Helper to report an exception traceback from its object
        """
        traceback = TracebackException.from_exception(exc)
        formatted = ''.join(traceback.format())
        log.error(formatted)

        if formatted in self._last_exceptions:
            log.debug('Exception already logged')
            return

        # Retain the formatted exception in memory to avoid looping
        self._last_exceptions.append(formatted)
        self._loop.call_later(
            self.EXCEPTION_TTL, self._forget_exception, formatted
        )

        self.send_report('exception', {'traceback': formatted})

    def _forget_exception(self, formatted):
        self._last_exceptions.remove(formatted)


sys.modules[__name__] = Reporter()



from aiohttp import web
from enum import Enum
import json
import logging

from nyuki.bus import reporting


log = logging.getLogger(__name__)
# aiohttp needs its own logger, and always prints HTTP hits using INFO level
access_log = logging.getLogger('.'.join([__name__, 'access']))
access_log.info = access_log.debug


class Method(Enum):

    """
    Supported HTTP methods by the REST API.
    """

    GET = 'GET'
    POST = 'POST'
    PUT = 'PUT'
    DELETE = 'DELETE'
    HEAD = 'HEAD'
    OPTIONS = 'OPTIONS'
    TRACE = 'TRACE'
    CONNECT = 'CONNECT'
    PATCH = 'PATCH'

    @classmethod
    def list(cls):
        return [method.name for method in cls]


class Response(web.Response):

    """
    Overrides aiohttp's response to facilitate its usage
    """

    ENCODING = 'utf-8'

    def __init__(self, body=None, **kwargs):

        # Check json
        if isinstance(body, dict) or isinstance(body, list):
            log.debug('converting dict/list response to bytes')
            body = json.dumps(body).encode(self.ENCODING)
            if not self._get_content_type(kwargs):
                kwargs['content_type'] = 'application/json'
        # Check body
        elif body is not None:
            log.debug('converting string response to bytes')
            body = str(body).encode(self.ENCODING)
            if not self._get_content_type(kwargs):
                kwargs['content_type'] = 'text/plain'

        return super().__init__(body=body, **kwargs)

    def _get_content_type(self, kwargs):
        return kwargs.get('content_type') or kwargs.get('headers', {}).get('content_type')


class Api(object):

    """
    The goal of this class is to gather all http-related behaviours.
    Basically, it's a wrapper around aiohttp.
    """

    def __init__(self, loop, debug=False, **kwargs):
        self._loop = loop
        self._server = None
        self._handler = None
        # Call order = list order
        self._middlewares = [mw_capability]
        self._debug = debug
        self._app = web.Application(
            loop=self._loop,
            middlewares=self._middlewares,
            **kwargs
        )

    @property
    def debug(self):
        return self._debug

    @debug.setter
    def debug(self, value):
        self._debug = bool(value)

    @property
    def router(self):
        return self._app.router

    async def build(self, host, port):
        """
        Create a HTTP server to expose the API.
        """
        log.info("Starting the http server on {}:{}".format(host, port))
        self._handler = self._app.make_handler(
            log=log, access_log=access_log, debug=self._debug
        )
        self._server = await self._loop.create_server(
            self._handler, host=host, port=port
        )

    async def destroy(self):
        """
        Gracefully destroy the HTTP server by closing all pending connections.
        """
        self._server.close()
        await self._handler.finish_connections()
        await self._server.wait_closed()
        log.info('Stopped the http server')


async def mw_capability(app, capa_handler):
    """
    Transform the request data to be passed through a capability and
    convert the result into a web response.
    """
    async def middleware(request):

        if request.method in request.POST_METHODS and hasattr(capa_handler, 'CONTENT_TYPE'):
            ctype = capa_handler.CONTENT_TYPE

            # Check content_type from @resource decorator
            if request.headers.get('Content-Type') != ctype:
                log.debug("content-type '%s' required", ctype)
                return Response(
                    {'error': 'Wrong content-type'},
                    status=400
                )

            # Check application/json is really a JSON body
            if ctype == 'application/json':
                try:
                    await request.json()
                except json.decoder.JSONDecodeError:
                    log.debug("request body for application/json must be JSON")
                    return Response(
                        {'error': 'application/json requires a JSON body'},
                        status=400
                    )

        try:
            capa_resp = await capa_handler(request, **request.match_info)
        except web.HTTPNotFound:
            # Avoid sending a report on a simple 404 Not Found
            raise
        except Exception as exc:
            reporting.exception(exc)
            raise exc

        if capa_resp and isinstance(capa_resp, Response):
            return capa_resp

        return Response()

    return middleware
